#include "basicnodes/basic.h"
#include "basicnodes/basic_factory.h"
#include "cfenodes/cfe.h"
#include "suifnodes/suif.h"
#include "suifpasses/passes.h"
#include "suifkernel/suif_env.h"
#include "suifkernel/module_subsystem.h"
#include "suifkernel/command_line_parsing.h"
#include "suifkernel/utilities.h"
#include "suifkernel/visitor_map.h"

LString k_proc_analyzed;
LString k_proc_being_analyzed;
LString k_const_proc;
LString k_pure_proc;
LString k_unreusable_proc;

bool is_auto(VariableSymbol *var);
bool is_addr_taken(VariableSymbol *var);
bool is_global(SymbolTable *st);
bool writes_volatile(StoreVariableStatement *stmt);

bool is_auto(VariableSymbol *var) {
  bool result = !is_global(var->get_symbol_table())
    && var->get_definition() == NULL;

  //fprintf(stderr, "var %s automatic: %d\n", var->get_name().c_str(), result);

  return result;
}

bool is_addr_taken(VariableSymbol *var) {
  return var->get_is_address_taken();
}

bool is_global(SymbolTable *st)
{
  if (st == NULL)
    return false;

  SuifObject *st_parent = st->get_parent();

  return is_kind_of<FileBlock>(st_parent)
    || is_kind_of<FileSetBlock>(st_parent);
}

bool writes_volatile(StoreVariableStatement *stmt) {
  VariableSymbol *dst = stmt->get_destination();
  
  if (!is_auto(dst) || is_addr_taken(dst)) {
    return true;
  }
   
  return false;
}

class ReuseAnalysisPass : public Pass {
private:
  void compute_reuse(ProcedureDefinition *pd);
  
public:
  ReuseAnalysisPass(SuifEnv *suif_env,
		    const LString &name = "reuse_analysis_pass");
  void initialize();

  Module *clone() const { return (Module *)this; } // is this correct?
  void do_file_set_block(FileSetBlock *file_set_block);
};

ReuseAnalysisPass::ReuseAnalysisPass(SuifEnv *suif_env, const LString &name) :
  Pass(suif_env, name)
{
}

void ReuseAnalysisPass::initialize() {
  Pass::initialize();

  k_proc_analyzed = "procedure analyzed for reuse";
  k_proc_being_analyzed = "procedure being analyzed for reuse";
  k_const_proc = "const";
  k_pure_proc = "pure";
  k_unreusable_proc = "unreusable procedure";
  
  /*
    OptionList *l;

    _command_line->set_description("analyze procedures for reusability");

    // collect optional flags
    OptionSelection *flags = new OptionSelection(false);
    l = new OptionList;
    l->add(new OptionLiteral("-input"));
    l->add(new OptionString("input file name"));
    l->set_description("input file");

    l = new OptionList;
    l->add(new OptionLiteral("-output"));
    l->add(new OptionString("output file name"));
    l->set_description("output file");

    _command_line->add(new OptionLoop(flags));
  */
}

void ReuseAnalysisPass::do_file_set_block(FileSetBlock *file_set_block) {
  for (Iter<FileBlock*> iter = file_set_block->get_file_block_iterator();
       iter.is_valid();
       iter.next()) {
    FileBlock *fb = iter.current();

    for (Iter<ProcedureDefinition*> iter = 
	   fb->get_definition_block()->get_procedure_definition_iterator();
	 iter.is_valid();
	 iter.next()) {
      ProcedureDefinition *procdef = iter.current();
      //fprintf(stderr, "procedure: %s\n", procdef->get_procedure_symbol()->get_name().c_str());
      compute_reuse(procdef);
      procdef->get_procedure_symbol()->append_annote(create_general_annote(get_suif_env(),
									   k_proc_analyzed));
    }
  }
}

void ReuseAnalysisPass::compute_reuse(ProcedureDefinition *pd) {
  if (pd->get_procedure_symbol()->peek_annote(k_proc_analyzed) || 
      pd->get_procedure_symbol()->peek_annote(k_unreusable_proc)) {
    return;
  }

  pd->get_procedure_symbol()->append_annote(create_general_annote(get_suif_env(),
								  k_proc_being_analyzed));

  bool is_cond_const = true;
  bool is_cond_pure = true;

  {
    list<StoreStatement*> *store_stmt_list = collect_objects<StoreStatement>(pd);

    /* not reusable if procedures contains StoreStatement(s) */
    if (!store_stmt_list->empty()) {
      //fprintf(stderr, "unreusable because of store statements\n");
      is_cond_const = false;
      is_cond_pure = false;
    }
    delete store_stmt_list;
  }    

  if (!is_cond_const && !is_cond_pure) {
    goto done;
  }

  {
    list<StoreVariableStatement*> *store_var_stmt_list = collect_objects<StoreVariableStatement>(pd);
  
    for (list<StoreVariableStatement*>::iterator iter = store_var_stmt_list->begin();
	 iter != store_var_stmt_list->end(); iter++) {
      StoreVariableStatement *store_var_stmt = *iter;
      VariableSymbol *stored_var = store_var_stmt->get_destination();

      if (!is_auto(stored_var) || is_addr_taken(stored_var)) {
	//fprintf(stderr, "unreusable because of store variable statements to %s\n", stored_var->get_name().c_str());
	is_cond_const = false;
	is_cond_pure = false;
	break;
      }
    }
    delete store_var_stmt_list;     
  }
 
  if (!is_cond_const && !is_cond_pure) {
    goto done;
  }

  {
    list<LoadVariableExpression*> *load_var_expr_list = collect_objects<LoadVariableExpression>(pd);
  
    for (list<LoadVariableExpression*>::iterator iter = load_var_expr_list->begin();
	 iter != load_var_expr_list->end(); iter++) {
      LoadVariableExpression *load_var_expr = *iter;
      VariableSymbol *loaded_var = load_var_expr->get_source();
      
      if (!is_auto(loaded_var) || is_addr_taken(loaded_var)) {
	is_cond_const = false;
	break;
      }
    }

    delete load_var_expr_list;
  }

  {
    list<LoadExpression*> *load_expr_list = collect_objects<LoadExpression>(pd);
  
    if (!load_expr_list->empty()) {
      is_cond_const = false;
    }

    delete load_expr_list;
  }

  {
    /* process procedure calls */
    list<CallStatement*> *call_stmt_list = collect_objects<CallStatement>(pd);

    for (list<CallStatement*>::iterator iter = call_stmt_list->begin();
	 iter != call_stmt_list->end(); iter++) {
      CallStatement *call_stmt = *iter;
    
      Expression *callee_address = call_stmt->get_callee_address();
      if (is_a<SymbolAddressExpression>(callee_address)) {
	SymbolAddressExpression *callee_symb_addr = to<SymbolAddressExpression>(callee_address);
	ProcedureSymbol *callee_proc_symb = to<ProcedureSymbol>(callee_symb_addr->get_addressed_symbol());

	if (callee_proc_symb->get_definition() == 0) {
	  //fprintf(stderr, "unreusable because of unknown procedure called\n");
	  is_cond_const = false;
	  is_cond_pure = false;
	  break;
	}
      
	if (!callee_proc_symb->peek_annote(k_proc_analyzed) &&
	    !callee_proc_symb->peek_annote(k_proc_being_analyzed)) {
	  compute_reuse(callee_proc_symb->get_definition());
	}
	
	if (callee_proc_symb->peek_annote(k_unreusable_proc)) {
	  //fprintf(stderr, "unreusable because of calling unreusable procedures\n");
	  is_cond_const = false;
	  is_cond_pure = false;
	  break;
	} else if (callee_proc_symb->peek_annote(k_pure_proc)) {
	  is_cond_const = false;
	  // don't break here
	}
      } else { // calling through non-symbols (pointers)
	//fprintf(stderr, "unreusable because of calling function pointers\n");
	is_cond_const = false;
	is_cond_pure = false;
	break;
      }
    }

    delete call_stmt_list;

    list<CallExpression*> *call_expr_list = collect_objects<CallExpression>(pd);
  
    for (list<CallExpression*>::iterator iter = call_expr_list->begin();
	 iter != call_expr_list->end(); iter++) {
      CallExpression *call_expr = *iter;
    
      Expression *callee_address = call_expr->get_callee_address();
      if (is_a<SymbolAddressExpression>(callee_address)) {
	SymbolAddressExpression *callee_symb_addr = to<SymbolAddressExpression>(callee_address);
	ProcedureSymbol *callee_proc_symb = to<ProcedureSymbol>(callee_symb_addr->get_addressed_symbol());

	if (callee_proc_symb->get_definition() == 0) {
	  //fprintf(stderr, "unreusable because of unknown procedure called\n");
	  is_cond_const = false;
	  is_cond_pure = false;
	  break;
	}
      
	if (!callee_proc_symb->peek_annote(k_proc_analyzed) &&
	    !callee_proc_symb->peek_annote(k_proc_being_analyzed)) {
	  compute_reuse(callee_proc_symb->get_definition());
	}
	
	if (callee_proc_symb->peek_annote(k_unreusable_proc)) {
	  //fprintf(stderr, "unreusable because of calling unreusable procedures\n");
	  is_cond_const = false;
	  is_cond_pure = false;
	  break;
	} else if (callee_proc_symb->peek_annote(k_pure_proc)) {
	  is_cond_const = false;
	  // don't break here
	}
      } else { // calling through non-symbol (pointers)
	//fprintf(stderr, "unreusable because of calling function pointers\n");
	is_cond_const = false;
	is_cond_pure = false;
	break;
      }
    }

    delete call_expr_list;
  }

 done:
  pd->get_procedure_symbol()->remove_annote_by_name(k_const_proc);
  pd->get_procedure_symbol()->remove_annote_by_name(k_pure_proc);
  pd->get_procedure_symbol()->remove_annote_by_name(k_unreusable_proc);

  if (is_cond_const) {
    //fprintf(stderr, "const procedure: %s\n", pd->get_procedure_symbol()->get_name().c_str());
    pd->get_procedure_symbol()->append_annote(create_general_annote(get_suif_env(),
								    k_const_proc));
  } else if (is_cond_pure) {
    //fprintf(stderr, "pure procedure: %s\n", pd->get_procedure_symbol()->get_name().c_str());
    pd->get_procedure_symbol()->append_annote(create_general_annote(get_suif_env(),
								    k_pure_proc));
  } else if (!is_cond_const && !is_cond_pure) {
    //fprintf(stderr, "unreusable procedure: %s\n", pd->get_procedure_symbol()->get_name().c_str());
    pd->get_procedure_symbol()->append_annote(create_general_annote(get_suif_env(),
								    k_unreusable_proc));
  }

  pd->get_procedure_symbol()->remove_annote_by_name(k_proc_being_analyzed);
}

extern "C" void EXPORT init_reuse_analysis(SuifEnv *suif_env) {
  suif_env->require_module("basicnodes");
  suif_env->require_module("cfenodes");
  suif_env->require_module("suifnodes");
  suif_env->require_module("suifpasses");
  suif_env->require_module("suifcloning");

  ModuleSubSystem *ms = suif_env->get_module_subsystem();
  ReuseAnalysisPass *pass = new ReuseAnalysisPass(suif_env);
  ms->register_module(pass);
}
